import asyncio
import warnings
from functools import wraps
from typing import Callable, TypeVar

import pydantic
from typing_extensions import ParamSpec

from souswift_core.models import Model

T = TypeVar('T')
P = ParamSpec('P')
ModelT = TypeVar('ModelT', bound=Model)


def deprecated(func: Callable[P, T]) -> Callable[P, T]:
    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        warnings.warn(
            f'{func.__name__} is deprecated and can be removed without notice!',
            DeprecationWarning,
        )
        return func(*args, **kwargs)

    return inner


async def as_async(
    func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    """Asynchronously schedule sync function to run in a separate thread"""
    return await asyncio.to_thread(func, *args, **kwargs)


def get_optional_model(model: type[ModelT], *exclude: str) -> type[ModelT]:
    edit_dto = pydantic.create_model(
        f'Optional{model.__name__}',
        __base__=model,
        __module__=model.__module__,
    )
    for field in edit_dto.__fields__.values():
        field.required = False
        field.default = None
    for field in exclude:
        edit_dto.__fields__.pop(field)
    return edit_dto



import abc
import importlib
import inspect
from pathlib import Path
from typing import Any, Generic, Optional, TypeVar


class _Finder(abc.ABC):
    def __init__(self, path: Path, root: Path = None):
        self.path = path
        self._targets = {}
        self.root = root or path

    def find(self):
        if self.path.is_dir():
            self.find_from_dir()
        else:
            self.find_from_file()

    def find_from_dir(self, path: Optional[Path] = None):
        if path is None:
            path = self.path
        if 'pycache' in path.name or '.pyc' in path.name:
            return
        for item in path.iterdir():
            if item.is_dir():
                self.find_from_dir(item)
                continue
            if '.py' in path.name and path.name != '__init__.py':
                continue
            self.find_from_file(item)

    def find_from_file(self, path: Optional[Path] = None):
        if path is None:
            path = self.path
        if not path.is_file():
            return
        mod = importlib.import_module(self.get_import(path))
        for name, obj in inspect.getmembers(mod):
            if self.is_valid_object(obj):
                self._targets[name] = obj

    def get_import(self, target: Path):
        target_str = target.as_posix()
        return (
            target_str.replace(self.root.as_posix(), self.root.name)
            .replace('/', '.')
            .replace('.py', '')
        )

    @abc.abstractmethod
    def is_valid_object(self, target: Any) -> bool:
        """Returns if :param:`target` meets the search criteria"""

    @abc.abstractmethod
    def __getitem__(self, name: str) -> Any:
        """Returns target by :param:`name`"""


T = TypeVar('T')


class InstanceFinder(_Finder, Generic[T]):
    def __init__(self, instance_of: type[T], path: Path, root: Path = None):
        self._instance_of = instance_of
        super().__init__(path, root=root)

    def is_valid_object(self, target: Any) -> bool:
        return isinstance(target, self._instance_of)

    def __getitem__(self, name: str) -> T:
        return self._targets[name]


class ClassFinder(_Finder, Generic[T]):
    def __init__(self, child_of: type[T], path: Path, root: Path = None):
        self._child_of = child_of
        super().__init__(path, root=root)

    def is_valid_object(self, target: Any) -> bool:
        if not inspect.isclass(target):
            return False
        is_child_of = issubclass(target, self._child_of)
        is_not_parent = target is not self._child_of
        return is_child_of and is_not_parent

    def __getitem__(self, name: str) -> type[T]:
        return self._targets[name]



from functools import partial
from http import HTTPStatus
from typing import Any, Callable

from fastapi import APIRouter
from fastapi.responses import ORJSONResponse


def _post(self, *args, **kwargs):
    _set_default_status(kwargs, HTTPStatus.CREATED)
    return APIRouter.post(self, *args, **kwargs)


def _delete(self, *args, **kwargs):
    _set_default_status(kwargs, HTTPStatus.NO_CONTENT)
    return APIRouter.delete(self, *args, **kwargs)


def _set_default_status(kwargs: dict[str, Any], status: int):
    kwargs.setdefault('status_code', status)


def _overload_function(target: APIRouter, name: str, func: Callable) -> None:
    setattr(target, name, partial(func, target))


def router_factory(create_post: bool, no_content_delete: bool, **router_args):
    """Returns a router with default response class as :class:`fastapi.responses.UJSONResponse`
    :param:`create_post` - overrides :method:`fastapi.APIRouter.post` to use HTTP_201_CREATED as default status code.
    :param:`no_content_delete` - overrides :method:`fastapi.APIRouter.delete` to use HTTP_204_NO_CONTENT as default status code.
    """
    router = APIRouter(default_response_class=ORJSONResponse, **router_args)
    if create_post:
        _overload_function(router, 'post', _post)
    if no_content_delete:
        _overload_function(router, 'delete', _delete)
    return router
